/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvpropertycontainer.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <QVDisjointSet>
#include <QVPropertyContainer>

uint QVPropertyContainer::maxIdent = 0;
// QVPropertyContainerInformer QVPropertyContainer::globalInformer;

QVPropertyContainer::QVPropertyContainer(const QString name):
	name(name), errorString(), variants(), safelyCopiedVariants(), minimum(),
	maximum(), _info(), io_flags(), link_flags(), insertion_order(),
	inputLinks(), outputLinks()
	{
	ident = getNewIdent();
	qDebug() << "QVPropertyContainer::QVPropertyContainer(" << name << ")";
	if(qvApp == NULL)
		{
		QString str = "QVPropertyContainer::QVPropertyContainer(): holder " + name +
					  ": property holders cannot be created before the " +
					  "QVApplication instance. Aborting now.";
		std::cerr << qPrintable(str) << std::endl;
		exit(1);
		}
	else
		{
		qvApp->registerQVPropertyContainer(this);
// 		QVPropertyContainer::globalInformer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::ContainerAdd));
		}
	qDebug() << "QVPropertyContainer::QVPropertyContainer(" << name << ") <- return";
	}

QVPropertyContainer::QVPropertyContainer(const QVPropertyContainer &cont):
	name(cont.name), ident(cont.ident), errorString(cont.errorString), variants(cont.variants),
	safelyCopiedVariants(cont.safelyCopiedVariants), minimum(cont.minimum), maximum(cont.maximum),
	_info(cont._info), io_flags(cont.io_flags), link_flags(cont.link_flags), insertion_order(cont.insertion_order),
	RWLock(), inputLinks(cont.inputLinks), outputLinks(cont.outputLinks) 
	{
	qDebug() << "QVPropertyContainer::QVPropertyContainer(" << name << ")";
	if(qvApp == NULL)
		{
		QString str = "QVPropertyContainer::QVPropertyContainer(): holder " + name +
					  ": property holders cannot be created before the " +
					  "QVApplication instance. Aborting now.";
		std::cerr << qPrintable(str) << std::endl;
		exit(1);
		}
	else
		{
		qvApp->registerQVPropertyContainer(this);
// 		QVPropertyContainer::globalInformer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::ContainerAdd));
		}
	qDebug() << "QVPropertyContainer::QVPropertyContainer(" << name << ") <- return";
	}

QVPropertyContainer & QVPropertyContainer::operator=(const QVPropertyContainer &cont)
	{
	name = cont.name;
	ident = cont.ident;
	errorString = cont.errorString;
	variants = cont.variants;
	safelyCopiedVariants = cont.safelyCopiedVariants;
	minimum = cont.minimum;
	maximum = cont.maximum;
	_info = cont._info;
	io_flags = cont.io_flags;
	link_flags = cont.link_flags;
	insertion_order = cont.insertion_order;
	inputLinks = cont.inputLinks;
	outputLinks = cont.outputLinks;

	informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::All));
	return *this;
	}

QVPropertyContainer::~QVPropertyContainer()
	{
	if(qvApp != NULL)
		qvApp->deregisterQVPropertyContainer(this);
	///@todo estudiar que pasa con estos unlink, al marcarse de esta manera como se destruye ya no se llama a read o write y creo que se queda el link creado como basura.
	unlink();
	informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::ContainerDel));
	}

void QVPropertyContainer::setName(const QString name)
	{ this->name = name; informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::Name)); }

const QString QVPropertyContainer::getName() const
	{ return this->name; }

const uint QVPropertyContainer::getId() const
	{ return this->ident; }

bool QVPropertyContainer::operator==(const QVPropertyContainer &cont) const
	{ return (ident == cont.ident); }

QList<QString> QVPropertyContainer::getPropertyList() const
	{ return variants.keys(); }

bool QVPropertyContainer::containsProperty(const QString name) const
	{ return variants.contains(name); }

int QVPropertyContainer::getPropertyType(const QString name, bool *ok) const
	{
	if(not checkExists(name,"QVPropertyContainer::getPropertyType()"))
		{
		if(ok != NULL) *ok = FALSE;
		return QVariant::Invalid;
		}
	if(ok != NULL) *ok = TRUE;
	QVariant variant = variants.value(name);
	return variant.userType();
	}

bool QVPropertyContainer::removeProperty(const QString name)
	{
	if(not checkExists(name,"QVPropertyContainer::removeProperty()"))
		return FALSE;
	this->variants.remove(name);
	this->safelyCopiedVariants.remove(name);
	this->minimum.remove(name);
	this->maximum.remove(name);
	this->_info.remove(name);
	this->io_flags.remove(name);
	int i = this->insertion_order.indexOf(name);
	this->insertion_order.removeAt(i);
	informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::PropertyDel, name));
	return TRUE;
	}
/*
bool QVPropertyContainer::setPropertyRange(const QString name, const double & minimum, const double & maximum)
	{
	std::cerr << "WARNING: setPropertyRange() is deprecated. Specify range in the function addProperty instead." << std::endl;
	if(not checkExists(name,"QVPropertyContainer::setPropertyRange()"))
		return FALSE;
	if(minimum <= getPropertyValue<double>(name) and
		maximum >= getPropertyValue<double>(name))
		{
			this->minimum[name] = QVariant::fromValue(minimum);
			this->maximum[name] = QVariant::fromValue(maximum);
			return TRUE;
		} else {
			QString str =  "QVPropertyContainer::setPropertyRange(): property " +
						   name + " in holder " + getName() + " has value " +
						   QString("%1").arg(getPropertyValue<double>(name)) +
						   ", which is not valid for the range [" +
						   QString("%1,%2").arg(minimum).arg(maximum) + "]." ;
			setLastError(str);
			if(qvApp->isRunning()) {
				std::cerr << qPrintable("Warning: " + str + "\n");
			} // Otherwise, qApp will show the error and won't start the program.
			return FALSE;
		}
	}

bool QVPropertyContainer::setPropertyRange(QString name, int & minimum, int & maximum)
	{ return setPropertyRange(name, static_cast<double>(minimum), static_cast<double>(maximum)); }
*/
bool QVPropertyContainer::hasRange(const QString name) const
	{ return maximum.contains(name) and minimum.contains(name); }

bool QVPropertyContainer::isInput(const QString name) const
	{ return (io_flags[name] & inputFlag);};

bool QVPropertyContainer::isOutput(const QString name) const
	{ return (io_flags[name] & outputFlag);};

bool QVPropertyContainer::isGUIInvisible(const QString name) const
	{ return (io_flags[name] & guiInvisible);};

bool QVPropertyContainer::isLinkedInput(const QString name) const
	{ return (link_flags[name] & linkedInputFlag);};

bool QVPropertyContainer::isLinkedOutput(const QString name) const
	{ return (link_flags[name] & linkedOutputFlag);};

QVariant QVPropertyContainer::getPropertyQVariantValue(const QString name, bool *ok) const
	{
	if (not checkExists(name,"QVPropertyContainer::getPropertyQVariantValue()"))
		if(ok != NULL) *ok = FALSE;
	else
		if(ok != NULL) *ok = TRUE;
	return variants[name];
	}

QString QVPropertyContainer::getPropertyInfo(const QString name, bool *ok) const
	{
	if(not checkExists(name,"QVPropertyContainer::getPropertyInfo()"))
		if(ok != NULL) *ok = FALSE;
	else
		if(ok != NULL) *ok = TRUE;
	return this->_info[name];
	}

QString QVPropertyContainer::getLastError() const
	{
	return errorString;
	}

const QString QVPropertyContainer::infoInputProperties() const
	{
	qDebug() << "QVPropertyContainer::infoInputProperties(" << getName() << ")";

	QString info = QString("Input parameters for ") + getName() + QString(":\n");
	bool emptyInfo=TRUE;

	qDebug() << "QVPropertyContainer::infoInputProperties(): Properties " << insertion_order;

	foreach (QString property, insertion_order)
	//QListIterator<QString> i(insertion_order);
	//while (i.hasNext())
		{
		//const QString property = i.next();

		if( not isInput(property) )
			continue;

		bool printableProperty = TRUE;
		QString propertyInfo("  --" + property + "=");

		switch(getPropertyType(property))
			{
			case QVariant::String:
				propertyInfo += QString() + "[text] " + "(def. '" + getPropertyValue<QString>(property) + "') ";
				break;

			case QVariant::Double:
				propertyInfo += ( (maximum.contains(property) and minimum.contains(property))?
					"[" + QString().setNum(getPropertyMinimum<double>(property)) + "..."
						+ QString().setNum(getPropertyMaximum<double>(property)) + "] ":
					"[double] " ) + "(def. "+ QString().setNum(getPropertyValue<double>(property)) + ") ";
				break;

			case QVariant::Int:
				propertyInfo += ( (maximum.contains(property) and minimum.contains(property))?
					"[" + QString().setNum(getPropertyMinimum<int>(property)) + "..." +
						QString().setNum(getPropertyMaximum<int>(property)) + "] ":
					"[int] " ) + "(def. "+ QString().setNum(getPropertyValue<int>(property)) + ") ";
				break;

			case QVariant::Bool:
				propertyInfo += "[true,false]" + (getPropertyValue<bool>(property) ?
								QString(" (def. true) "):QString("(def. false) "));
				break;

			default:
				printableProperty = FALSE;
				break;
			}

		if (printableProperty)
			{
			info += propertyInfo + getPropertyInfo(property).rightJustified(100-propertyInfo.split('\n').last().length(),' ') + ".\n";
			emptyInfo=FALSE;
			}
		}

	qDebug() << "QVPropertyContainer::infoInputProperties(" << getName() << ") <~ return";

	if(emptyInfo)
		return QString("");
	
	return info;
	}

bool QVPropertyContainer::correctRange(const QString name, const double & value) const
	{
	if(not maximum.contains(name) and not minimum.contains(name))
		return TRUE;
	double maximum = getPropertyMaximum<double>(name);
	double minimum = getPropertyMinimum<double>(name);
	if(minimum <= value and maximum >= value)
		return TRUE;
	else
		{
		QString str =  "QVPropertyContainer::setPropertyValue(): value " +
					   QString("%1").arg(value) + " for property " +
					   name + " in holder " + getName() +
					   "is not valid for the range [" +
					   QString("%1,%2").arg(minimum).arg(maximum) + 
					   "] stablished for it." ;
		setLastError(str);
		if(qvApp->isRunning())
			{
			std::cerr << qPrintable("Warning: " + str + "\n");
			} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		}
	}

bool QVPropertyContainer::correctRange(const char *name, const int & value) const
	{ return correctRange(QString(name),static_cast<double>(value)); }

bool QVPropertyContainer::correctRange(QString name, const int & value) const
	{ return correctRange(name,static_cast<double>(value)); }

bool QVPropertyContainer::checkExists(const QString name, const QString methodname) const
	{
	if(not variants.contains(name))
		{
		QString str =  methodname + ": property " + name +
					   " doesn't exists in holder " + getName() + ".";
		setLastError(str);
		if(qvApp->isRunning()) {
			std::cerr << qPrintable("Warning: " + str + "\n");
		} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		} else {
		return TRUE;
		}
	}

bool QVPropertyContainer::checkIsNewProperty(const QString name, const QString methodname) const
	{
	if(variants.contains(name))
		{
		QString str =  methodname + "(): property " + name +
					   " already exists in holder " + getName() + ".";
		setLastError(str);
		if(qvApp->isRunning()) {
			std::cerr << qPrintable("Warning: " + str + "\n");
		} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		} else {
		return TRUE;
		}
	}

bool QVPropertyContainer::linkProperty(QString prop_orig, QVPropertyContainer *qvp_dest, QString prop_dest, LinkType link_type)
	{
	bool ok1,ok2;
	QString errMsg;
	int t1,t2;
	QVPropertyContainer *qvp_err=NULL;
	
	t1 = this->getPropertyType(prop_orig,&ok1);
	t2 = qvp_dest->getPropertyType(prop_dest,&ok2);
	if(qvApp->isRunning())
		{
		qvp_err = this;
		errMsg = QString("QVPropertyContainer::linkProperty(): Property holder %1:"
					"Cannot link properties after launching QVApplication.\n")
					.arg(prop_orig).arg(this->getName());
		}
	else if(this == qvp_dest)
		{
		errMsg = QString("QVPropertyContainer::linkProperty(): Property holder %1: cannot link a QVPropertyContainer with itself.\n").arg(this->getName());
		qvp_err = this;
		}
	else if(not ok1)
		{
		errMsg = QString("QVPropertyContainer::linkProperty(): Property %1 does not exist in property holder %2.\n")
				.arg(prop_orig).arg(this->getName());
		qvp_err = this;
		}
	else if (not ok2)
		{
		errMsg = QString("QVPropertyContainer::linkProperty(): Property %1 does not exist in property holder %2.\n")
				.arg(prop_dest).arg(qvp_dest->getName());
		qvp_err = qvp_dest;
		}
	else if(t1 != t2)
		{
		errMsg = QString("QVPropertyContainer::linkProperty(): Properties %1 and %2 of QVPropertyContainers %3 and %4 respectively are not of the same type.\n").arg(prop_orig).arg(prop_dest).arg(this->getName()).arg(qvp_dest->getName());
		qvp_err = this;
		}
	else if(not (this->io_flags[prop_orig] & outputFlag))
		{
		errMsg = QString("QVPropertyContainer::linkProperty(): Property %1 of property holder %2 is not of Output type, and cannot be linked as such.\n").arg(prop_orig).arg(this->getName());
		qvp_err = this;
		}
	else if(not (qvp_dest->io_flags[prop_dest] & inputFlag))
		{
		errMsg = QString("QVPropertyContainer::linkProperty(): Property %1 property holder %2 is not of Input type, and cannot be linked as such.\n").arg(prop_dest).arg(qvp_dest->getName());
		qvp_err = qvp_dest;
		}

	if(errMsg != QString())
		{
		qvp_err->setLastError(errMsg);
		if(qvApp->isRunning()) {
			std::cerr << qPrintable("Warning: " + errMsg + "\n");
		} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		}
	else
		{
		QVPropertyContainerLink *link = new QVPropertyContainerLink(this,prop_orig,qvp_dest,prop_dest,link_type);
		this->outputLinks[prop_orig].push_back(link);
		this->link_flags[prop_orig] |= linkedOutputFlag;
		// We add a new metod "addInputLink(prop_dest, link);" instead of
		// do: "qvp_dest->inputLinks[prop_dest] = link; qvp_dest->link_flags[prop_dest] |= linkedInputFlag;", in order to
		// qvp_dest can emit a signal saying that its links were changed
		qvp_dest->addInputLink(prop_dest, link);

		// Now, we initialize the shared state, simply protected by the
		// corresponding RWLock:
		this->RWLock.lockForWrite();
		safelyCopiedVariants[prop_orig] = variants[prop_orig];
		this->RWLock.unlock();

		if (link_type == QVPropertyContainer::AsynchronousLink)
			informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::LinkAdd, this->getName(), prop_orig, qvp_dest->getName(), prop_dest,FALSE));
		else
			informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::LinkAdd, this->getName(), prop_orig, qvp_dest->getName(), prop_dest,TRUE));
		return TRUE;
		}
	}

void QVPropertyContainer::addInputLink(QString prop_dest, QVPropertyContainerLink *link)
	{
		inputLinks[prop_dest] = link;
		link_flags[prop_dest] |= linkedInputFlag;
	}

void QVPropertyContainer::linkProperty(QVPropertyContainer *container, LinkType linkType)
	{
	QList<QString> localProper = getPropertyList();
	QList<QString> inputProper = container->getPropertyList();

	qSort(localProper.begin(), localProper.end());
	qSort(inputProper.begin(), inputProper.end());

	int i=0, l=0;
	while ( (i < inputProper.size()) && (l < localProper.size()) )
		{
		QString localName = localProper[l];
		QString inputName = inputProper[i];
		if (
			(localName == inputName) && // si tienen el mismo nombre
			(getPropertyType(localName) == container->getPropertyType(inputName)) && // si tienen el mismo tipo
			(io_flags[localName] & outputFlag) && // si una es de entrada
			(container->io_flags[inputName] & inputFlag) && // y otra de salida
			(!(io_flags[localName] & internalProp)) && // y no son propiedades internas
			(!(container->io_flags[inputName] & internalProp))
		   )
			linkProperty(localProper[l], container, inputProper[i], linkType);

		if(localName <= inputName) l++;
		if(localName >= inputName) i++;
		}
	}

bool QVPropertyContainer::unlinkProperty(QString origName, QVPropertyContainer *destCont, QString destName)
	{
		QList<QVPropertyContainerLink*> linkList = outputLinks[origName];
		foreach(QVPropertyContainerLink* link, linkList) {
			if ( (link->qvp_orig == this) && (link->prop_orig == origName) &&
				(link->qvp_dest == destCont) && (link->prop_dest == destName) ) {
				link->markedForDeletion = TRUE;
				// Protect against a possible pending acquire() for our output
				// in other holders:
				link->SyncSemaphoreOut.release();
				informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::LinkDel, this->getName(), origName, destCont->getName(), destName));
				return TRUE;
			}
		}
		return FALSE;
	}

void QVPropertyContainer::unlink()
{
	foreach(QVPropertyContainerLink* link, inputLinks.values()) {
		link->markedForDeletion = TRUE;
		// Protect against a possible pending acquire() from our input
		// in other holders:
		link->SyncSemaphoreIn.release();
		// using the orig and dest names because references should be bads if we are unlink all the items
		informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::LinkDel, link->qvp_orig_name, link->prop_orig, link->qvp_dest_name, link->prop_dest));
	}

	foreach(QList<QVPropertyContainerLink*> linkList, outputLinks.values()) {
		foreach(QVPropertyContainerLink* link, linkList) {
			link->markedForDeletion = TRUE;
			// Protect against a possible pending acquire() for our output
			// in other holders:
			link->SyncSemaphoreOut.release();
			// using the orig and dest names because references should be bads if we are unlink all the items
			informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::LinkDel, link->qvp_orig_name, link->prop_orig, link->qvp_dest_name, link->prop_dest));
		}
	}
}

void QVPropertyContainer::readInputProperties()
	{
	// We read every linked input property from its source, protecting
	// the read with a standard RWLock (that here we just lock for read).
	// The only caveat is that if a property is synchronously read, then
	// we must wait for the producer to write it first. We implement that
	// by waiting on the SyncSemaphoreOut of the link. Also, in this
	// case when we finish reading the property, we signal every possible
	// waiting writer that, regarding this specific link, it can write now
	// a new value if it needs to, because we have read the old value yet.
	// This is implemented by releasing the SyncSemaphoreIn associated to
	// the link.
	QMutableMapIterator<QString, QVPropertyContainerLink*> i(inputLinks);
	while (i.hasNext()) {
		i.next();
		QVPropertyContainerLink *link = i.value();
		if(link->link_type == SynchronousLink and not link->markedForDeletion) {
			link->SyncSemaphoreOut.acquire();
		}
		link->qvp_orig->RWLock.lockForRead();
		//this->setPropertyValueQVariant(link->prop_dest,link->qvp_orig->safelyCopiedVariants[link->prop_orig]);
		this->variants[link->prop_dest] = link->qvp_orig->safelyCopiedVariants[link->prop_orig];
		link->qvp_orig->RWLock.unlock();
		if(link->link_type == SynchronousLink and not link->markedForDeletion) {
			link->SyncSemaphoreIn.release();
		}
		informer.emitChange(QVPropertyContainerChange(this, QVPropertyContainerChange::PropertyValue, link->prop_dest));
		// Possible link deletion:
		if(link->markedForDeletion) {
			i.remove();
			toDeleteLink(link);
		}
	}
}

void QVPropertyContainer::writeOutputProperties()
	{
	QMutableMapIterator<QString, QList<QVPropertyContainerLink*> >i(outputLinks);

	// For every QVP synchronously linked to this QVP's output, we ask
	// for permision to write a new output (that will only be possible if
	// all of these QVP's have read their inputs already):
	while (i.hasNext()) {
		i.next();
		QListIterator<QVPropertyContainerLink*> j(i.value());
		while(j.hasNext()) {
			QVPropertyContainerLink *link = j.next();
			if(link->link_type == SynchronousLink and not link->markedForDeletion) {
				link->SyncSemaphoreIn.acquire();
			}
		}
	}

	// Now, we write a new coherent state, simply protected by the
	// corresponding RWLock:
	i.toFront();
	this->RWLock.lockForWrite();
	while (i.hasNext()) {
		i.next();
		QString prop_orig = i.key();
		safelyCopiedVariants[prop_orig] = variants[prop_orig];
	}
	this->RWLock.unlock();

	// Finally, we signal to QVP's synchronously linked to this QVP's output
	// that there is a new coherent output, by releasing the corresponding
	// semaphore.
	i.toFront();
	while (i.hasNext()) {
		i.next();
		QMutableListIterator<QVPropertyContainerLink*> j(i.value());
		while(j.hasNext()) {
			QVPropertyContainerLink *link = j.next();
			if(link->link_type == SynchronousLink and not link->markedForDeletion) {
				link->SyncSemaphoreOut.release();
			}
			// Possible link deletion:
			if(link->markedForDeletion) {
				j.remove();
				toDeleteLink(link);
				if(i.value().isEmpty()) {
					i.remove();
					break;
				}
			}
		}
	}
}

void QVPropertyContainer::toDeleteLink(QVPropertyContainerLink* link)
	{
		if (link->qvp_orig == this) {
			link->qvp_orig = NULL;
		}
		else if (link->qvp_dest == this) {
			link->qvp_dest = NULL;
		}

		if ((link->qvp_orig == NULL) && (link->qvp_dest == NULL)) delete link;
	}

void QVPropertyContainer::setLastError(QString str) const
	{ errorString = str; }

bool QVPropertyContainer::areSynchronized(const QList<QVPropertyContainer *> conts)
	{
	QVDisjointSet dSet(conts.size());

	for (int i = 0; i < conts.size(); i++)
		for (int j = i+1; j < conts.size(); j++) // for each pair of container
			{
			bool find = false; // only need one syncrhonous property to be synchronized

			if (conts.at(i)->getId() == conts.at(j)->getId()) // if they are the same container, they are synchronized
				{
				dSet.unify(i, j); // unify their sets
				find = true;
				}
 			if (!find)
				{
				const QMap<QString, QVPropertyContainerLink* > inLinksI = conts.at(i)->getInputLinks();
				foreach(QVPropertyContainerLink* proConLink, inLinksI.values())        // for each first container's linked input property
					if ( (proConLink->qvp_orig->getId() == conts.at(j)->getId()) &&   // if is linked from second container
					     (proConLink->link_type == SynchronousLink)                ) // and is a synchronous link
						{
						dSet.unify(i, j); // unify their sets
						find = true;
						break;
						}
				}
 			if (!find)
				{
				const QMap<QString, QVPropertyContainerLink* > inLinksJ = conts.at(j)->getInputLinks();
				foreach(QVPropertyContainerLink* proConLink, inLinksJ.values())        // for each second container's linked input property
 					if ( (proConLink->qvp_orig->getId() == conts.at(i)->getId()) &&   // if is linked from first container
					     (proConLink->link_type == SynchronousLink)                ) // and is a synchronous link
 						{
 						dSet.unify(i, j); // unify their sets
 						break;
 						}
				}
			}

	return (dSet.numberOfSets() == 1);
	}

template <> bool QVPropertyContainer::parseArgument<bool>(const QString parameter, const QString value)
	{
	if (value.toLower() == "true" || value.toLower() == "false")
		{
		//variants[parameter] = QVariant::fromValue<bool>(value.toLower() == "true");
		setPropertyValue<bool>(parameter,value.toLower() == "true");
		return TRUE;
		}
	else {
		errorString = "QVPropertyContainer::parseArgument(): holder " + getName() +
					": value " + value +
					" is not a valid boolean value for parameter " +
					parameter + ".\n";
		return FALSE;
		}
	}

template <> bool QVPropertyContainer::parseArgument<int>(const QString parameter, const QString value)
	{
	bool okInt;
	int intValue = value.toInt(&okInt);
	if(not okInt)
		{
		errorString = "QVPropertyContainer::parseArgument(): holder " + getName() +
					": value " + value +
					" is not a valid integer value for parameter " +
					parameter + ".\n";
		return FALSE;
		}
	//variants[parameter] = QVariant::fromValue<int>(intValue);
	setPropertyValue<int>(parameter,intValue);
	return TRUE;
	}

template <> bool QVPropertyContainer::parseArgument<double>(const QString parameter, const QString value)
	{
	bool okDouble;
	double doubleValue = value.toDouble(&okDouble);
	if(not okDouble)
		{
		errorString = "QVPropertyContainer::parseArgument(): holder " + getName() +
					": value " + value +
					" is not a valid double value for parameter " +
					parameter + ".\n";
		return FALSE;
		}
	//variants[parameter] = QVariant::fromValue<double>(doubleValue);
	setPropertyValue<double>(parameter,doubleValue);
	return TRUE;
	}

template <> bool QVPropertyContainer::parseArgument<QString>(const QString parameter, const QString value)
	{
	//variants[parameter] = QVariant::fromValue<QString>(value);
	setPropertyValue<QString>(parameter,value);
	return TRUE;
	}
