/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

#include <qvprojective.h>
#include <qvmatrixalgebra.h>
#include <qvdefines.h>
#include <qvmath.h>
#include <float.h>

/// @file qvprojective.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

QVMatrix ComputeProjectiveHomography(const QList< QPair<QPointF, QPointF> > &matchings)
	{
	Q_ASSERT(matchings.size() >= 4);

	const QList<QPointF>	sourcePoints = getFirstPairList<QPointF, QPointF>(matchings),
				destPoints = getSecondPairList<QPointF, QPointF>(matchings);

	const int num_points = sourcePoints.size();

	// 0. Fast normalization (similar to DLT, but based on min and max values
	// of X and Y, instead of mean value and normalized mean distance to
	// centroid equal to sqrt(2). Hartley-Zisserman 2nd ed. pag 109.
	float minXS=FLT_MAX, maxXS=-FLT_MAX, minYS=FLT_MAX, maxYS=-FLT_MAX,
	      minXD=FLT_MAX, maxXD=-FLT_MAX, minYD=FLT_MAX, maxYD=-FLT_MAX;
	for (int i=0; i<num_points; i++)
		{
			if(sourcePoints[i].x() < minXS)
				minXS = sourcePoints[i].x();
			if(sourcePoints[i].x() > maxXS)
				maxXS = sourcePoints[i].x();
			if(sourcePoints[i].y() < minYS)
				minYS = sourcePoints[i].y();
			if(sourcePoints[i].y() > maxYS)
				maxYS = sourcePoints[i].y();
			if(destPoints[i].x() < minXD)
				minXD = destPoints[i].x();
			if(destPoints[i].x() > maxXD)
				maxXD = destPoints[i].x();
			if(destPoints[i].y() < minYD)
				minYD = destPoints[i].y();
			if(destPoints[i].y() > maxYD)
				maxYD = destPoints[i].y();
		}
	// This is to avoid possible div by zero in degenerate conditions:
	if(fabs(minXS-maxXS) < EPSILON)
		maxXS += 1.0;
	if(fabs(minYS-maxYS) < EPSILON)
		maxYS += 1.0;
	if(fabs(minXD-maxXD) < EPSILON)
		maxXD += 1.0;
	if(fabs(minYD-maxYD) < EPSILON)
		maxYD += 1.0;

	// 1. Obtain coeficient matrix for the system:
	//	Ax = 0
	QVMatrix coefsMatrix(3*num_points,9);
	for (int n=0;n<num_points;n++)
		{
		double x = (sourcePoints[n].x()-(maxXS+minXS)/2)/(maxXS-minXS),
		       y = (sourcePoints[n].y()-(maxYS+minYS)/2)/(maxYS-minYS),
		       x_p = (destPoints[n].x()-(maxXD+minXD)/2)/(maxXD-minXD),
		       y_p = (destPoints[n].y()-(maxYD+minYD)/2)/(maxYD-minYD);

		coefsMatrix(3*n, 0) = 0;	coefsMatrix(3*n, 1) = 0; 	coefsMatrix(3*n, 2) = 0;
		coefsMatrix(3*n, 3) = -x;	coefsMatrix(3*n, 4) = -y; 	coefsMatrix(3*n, 5) = -1;
		coefsMatrix(3*n, 6) = x*y_p;	coefsMatrix(3*n, 7) = y*y_p; 	coefsMatrix(3*n, 8) = y_p;

		coefsMatrix(3*n+1, 0) = x;	coefsMatrix(3*n+1, 1) = y; 	coefsMatrix(3*n+1, 2) = 1;
		coefsMatrix(3*n+1, 3) = 0;	coefsMatrix(3*n+1, 4) = 0; 	coefsMatrix(3*n+1, 5) = 0;
		coefsMatrix(3*n+1, 6) = -x*x_p;	coefsMatrix(3*n+1, 7) = -y*x_p; coefsMatrix(3*n+1, 8) = -x_p;

		coefsMatrix(3*n+2, 0) = -x*y_p;	coefsMatrix(3*n+2, 1) = -y*y_p; coefsMatrix(3*n+2, 2) = -y_p;
		coefsMatrix(3*n+2, 3) = x*x_p;	coefsMatrix(3*n+2, 4) = y*x_p; 	coefsMatrix(3*n+2, 5) = x_p;
		coefsMatrix(3*n+2, 6) = 0;	coefsMatrix(3*n+2, 7) = 0; 	coefsMatrix(3*n+2, 8) = 0;
		}

	// 2. Solve the homogeneous system.
	QVVector x(9);
	solveHomogeneousLinear(coefsMatrix, x);
	//solveHomogeneousLinear2(coefsMatrix, x);

	//x = x / x.norm2();
	//x2 = x2 / x2.norm2();
	//if (x[0] * x2[0] < 0)
	//	x2 = x2 *(-1);

	//if (x2[1] == 0)
	/*	{
		std::cout << "-----" << std::endl;
		std::cout << "x1 = " << x << std::endl;
		std::cout << "x2 = " << x2 << std::endl;
		}*/

	//x = x / x[8];

	// 3. Rearrange elements of vector 'x' in a 3x3 matrix.
	QVMatrix homography(3,3);
	homography(0,0) = x[0];	homography(0,1) = x[1];	homography(0,2) = x[2];
	homography(1,0) = x[3];	homography(1,1) = x[4];	homography(1,2) = x[5];
	homography(2,0) = x[6];	homography(2,1) = x[7];	homography(2,2) = x[8];

	//////////////////////////////////////////////////////////
	// 4. Renormalize homography back:
	QVMatrix premult(3,3),postmult(3,3);
	premult(0,0) = 1/(maxXS-minXS);	premult(0,1) = 0;		premult(0,2) = -(maxXS+minXS)/(2*(maxXS-minXS));
	premult(1,0) = 0;		premult(1,1) = 1/(maxYS-minYS);	premult(1,2) = -(maxYS+minYS)/(2*(maxYS-minYS));
	premult(2,0) = 0;		premult(2,1) = 0;		premult(2,2) = 1;

	postmult(0,0) = maxXD-minXD;	postmult(0,1) = 0;		postmult(0,2) = (maxXD+minXD)/2;
	postmult(1,0) = 0;		postmult(1,1) = maxYD-minYD;	postmult(1,2) = (maxYD+minYD)/2;
	postmult(2,0) = 0; 		postmult(2,1) = 0; 		postmult(2,2) = 1;

	homography = (postmult * homography) * premult;

	/*float normalizer=homography(2,2);
	for(int i=0;i<3;i++)
		for(int j=0;j<3;j++)
			homography(i,j) /= normalizer;*/

	return homography;
	}

QVMatrix ComputeHomography(const QList<QPointF> &sourcePoints, const QList<QPointF> &destPoints)
	{
	Q_ASSERT(sourcePoints.size() == destPoints.size());
	Q_ASSERT(sourcePoints.size() >= 4);

	std::cerr << "DEPRECATED: ComputeHomography is deprecated. use ComputeProjectiveHomography instead." << std::endl;
	return ComputeProjectiveHomography(joinPairList<QPointF, QPointF>(sourcePoints, destPoints));
	}

QVMatrix ComputeEuclideanHomography(const QPair<QPointF, QPointF> &firstMatching, const QPair<QPointF, QPointF> &secondMatching)
	{
	std::cerr << "DEPRECATED: This version of ComputeEuclideanHomography is deprecated. Use the other overloaded version instead, which only takes a list of matchings as input." << std::endl;
	const QPointF	A1 = firstMatching.first, B1 = secondMatching.first, A2 = firstMatching.second, B2 = secondMatching.second;
	const QPointF	C1 = (A1 + B1) / 2, C2 = (A2 + B2) / 2;
	const double	zoom = norm2(A2-B2) / norm2(A1-B1),
			delta = qvClockWiseAngle(A1-B1, A2-B2);

	QVMatrix result = QVMatrix::identity(3);
	result(0,0) = zoom*cos(delta);	result(0,1) = zoom*sin(delta);	result(0,2) = C2.x() - zoom*cos(delta)*C1.x() - zoom*C1.y()*sin(delta);
	result(1,0) = -zoom*sin(delta);	result(1,1) = zoom*cos(delta);	result(1,2) = C2.y() - zoom*cos(delta)*C1.y() + zoom*C1.x()*sin(delta);

	return result;
	}

void getMeanDirection(const QVMatrix m, QVVector &mean, QVVector &direction)
	{
	mean = m.meanCol();
	QVMatrix centeredM = m;
	for (int i = 0; i < centeredM.getCols(); i++)
		centeredM.setCol(i, centeredM.getCol(i) - mean);

	// 1. Compute main direction vector.
	QVMatrix eigVecs;
	QVVector eigVals;
	eigenDecomposition(centeredM * centeredM.transpose(), eigVals, eigVecs);

	direction = QVVector(eigVecs.getCols());
	for (int i = 0; i < eigVecs.getCols(); i++)
		direction[i] = eigVecs(0,i);
	direction = direction * eigVals[0];
	}


QVMatrix ComputeEuclideanHomography(const QList< QPair<QPointF, QPointF> > &matchings)
	{
	// 1. Get source and destination mean point and direction.
	const QList<QPointF> sourcePointList = getFirstPairList<QPointF, QPointF>(matchings),
				destPointList = getSecondPairList<QPointF, QPointF>(matchings);

	const QVMatrix sources = sourcePointList, destinations = destPointList;

	QVVector sourcesMean, destinationsMean, sourcesDirection, destinationsDirection;
	getMeanDirection(sources.transpose(), sourcesMean, sourcesDirection);
	getMeanDirection(destinations.transpose(), destinationsMean, destinationsDirection);

	const QPointF C1 = sourcesMean, C2 = destinationsMean, D1 = sourcesDirection, D2 = destinationsDirection;

	// 2. Get zoom and angle
	double zoom = 0;
	int switchD1Direction = 0;
	for(int i = 0; i < sourcePointList.size(); i ++)
		{
		const QPointF sourceCenteredPoint = sourcePointList.at(i) - sourcesMean, destCenteredPoint = destPointList.at(i) - destinationsMean;
		zoom += norm2(destCenteredPoint) / norm2(sourceCenteredPoint);

		if (	(norm2(destCenteredPoint - D2) - norm2(destCenteredPoint + D2)) * (norm2(sourceCenteredPoint + D1) - norm2(sourceCenteredPoint - D1)) > 0	|| 
			(norm2(destCenteredPoint - D2) - norm2(destCenteredPoint + D2)) * (norm2(sourceCenteredPoint + D1) - norm2(sourceCenteredPoint - D1)) > 0	)
			switchD1Direction++;
		else
			switchD1Direction--;
		}

	zoom /= sourcePointList.size();

	const double delta = qvClockWiseAngle((switchD1Direction > 0)?-D1:D1, D2);

	// 3. Get the euclidean homography matrix, derived from the following code for Mathematica:
	//	Rotation[delta_] := { {Cos[delta], Sin[delta], 0}, {-Sin[delta], Cos[delta], 0}, {0, 0, 1} };
	//	Translation[x_, y_] := { {1, 0, x}, {0, 1, y}, {0, 0, 1} };
	//	Zoom[z_] := { {z, 0, 0}, {0, z, 0}, {0, 0, 1} };
	//	result = Translation[C2_x, C2_y].Zoom[zoom].Rotation[delta].Translation[C1_x, C1_y] // MatrixForm
	QVMatrix result = QVMatrix::identity(3);
	result(0,0) = zoom*cos(delta);	result(0,1) = zoom*sin(delta);	result(0,2) = C2.x() - zoom*cos(delta)*C1.x() - zoom*C1.y()*sin(delta);
	result(1,0) = -zoom*sin(delta);	result(1,1) = zoom*cos(delta);	result(1,2) = C2.y() - zoom*cos(delta)*C1.y() + zoom*C1.x()*sin(delta);

	return result;
	}

QPointF ApplyHomography(const QVMatrix &H, const QPointF &point)
	{
	const double	h11 = H(0,0), h12 = H(0,1), h13 = H(0,2),
			h21 = H(1,0), h22 = H(1,1), h23 = H(1,2),
			h31 = H(2,0), h32 = H(2,1), h33 = H(2,2),
			x = point.x(), y = point.y(),
			homogenizer = h31*x + h32*y + h33;

	return QPointF(h11*x + h12*y + h13, h21*x + h22*y + h23)/homogenizer;
	}

QList<QPointF> ApplyHomography(const QVMatrix &homography, const QList<QPointF> &sourcePoints)
	{
	QList<QPointF> result;
	foreach(QPointF point, sourcePoints)
		result.append(ApplyHomography(homography, point));
	return result;
	}

double HomographyTestError(const QVMatrix &homography)
	{
	const QVVector v1 = homography.getCol(0), v2 = homography.getCol(1);
	return	ABS(v1.norm2() - v2.norm2()) / (v1.norm2() + v2.norm2())
		+ ABS(v1 * v2) / (v1.norm2() * v2.norm2());
	}



////////////////////////////////////////////////////////////////
// Focal calibration from homography
//
#include <qvmath/qvfunctionminimizer.h>
#ifndef DOXYGEN_IGNORE_THIS
class QVFocalCalibration: public QVFunctionMinimizer
	{
	private:
		QVMatrix H;

	public:	
		QVFocalCalibration(const QVMatrix &Horig): QVFunctionMinimizer(true), H(pseudoInverse(Horig))	{}

		virtual double function(const double value)
			{
			// Error function, given an 'f', and a 'H' matrix. In Mathematica:
			//	K = {{f, 0, 0}, {0, f, 0}, {0, 0, 1}};
			//	Hvar = {{h1, h2, h3}, {h4, h5, h6}, {h7, h8, h9}};
			//	ErrorMat[H_, K_] := Transpose[H].Inverse[K].Inverse[K].H;
			//	Error [ErrorM_] := ((ErrorM[[1, 1]] - ErrorM[[2, 2]])^2 + ErrorM[[1, 2]]^ 2) / ErrorM[[1, 1]]^ 2;
			//	ErrorFunction[H_, K_] := Error[ErrorMat[H, K]];
			//	ErrorFunction[Hvar, K] // Simplify
			const double	*h = H.getWriteData(), f = value;
			const double	h1 = h[0*3+0],	h2 = h[0*3+1],
					h4 = h[1*3+0],	h5 = h[1*3+1],
					h7 = h[2*3+0],	h8 = h[2*3+1];
		
			return		( POW2( h1*h2 + h4*h5 + (f)*(f)*h7*h8 ) + POW2( h1*h1 - h2*h2 + h4*h4 - h5*h5 + (f)*(f)*(h7*h7-h8*h8) ) )
					/ POW2(h1*h1 + h4*h4 + (f)*(f)*h7*h7);
			}
	};
#endif

void GetIntrinsicCameraMatrixFromHomography(const QVMatrix &H, QVMatrix &K, double focal, const double maxFocal, const int maxIterations, const double maxError)
	{
	QVFocalCalibration minimizer(H);

	minimizer.setStartingValue(focal);
	minimizer.setSearchRange(-maxFocal, maxFocal);

	const int status = minimizer.iterate(maxIterations, maxError);

	focal = minimizer.getMinimum();

	std::cout << "Focal = " << focal << std::endl;

	Q_ASSERT_X(status == GSL_SUCCESS, "GetIntrinsicCameraMatrixFromHomography", "Minimization method didn't converged: " + status);

	K = QVMatrix::identity(3);
	K(2,2) = 1/focal;
	}

// To get direct focal distance formula:
//
//	K = {{f, 0, 0}, {0, f, 0}, {0, 0, 1}};
//	Hvar = {{h1, h2, h3}, {h4, h5, h6}, {h7, h8, h9}};
//	ErrorMat[H_, K_] := Transpose[H].Inverse[K].Inverse[K].H;
//	Error[ErrorM_] := ((ErrorM[[1, 1]] - ErrorM[[2, 2]])^2 + ErrorM[[1, 2]]^2) / ErrorM[[1, 1]]^2;
//	ErrorFunction[H_, K_] := Error[ErrorMat[H, K]];
//	P = Dt[ErrorFunction[Hvar, K], f, Constants -> {h1, h2, h3, h4, h5, h6, h7, h8, h9}] ;
//	focalDistance  = Solve[P == 0, f][[2]][[1]] // FullSimplify
//	CForm[focalDistance]
//
void GetDirectIntrinsicCameraMatrixFromHomography(const QVMatrix &H, QVMatrix &K)
	{
	const double *h = H.getReadData();
	const double h1 = h[0*3+0], h2 = h[0*3+1], h4 = h[1*3+0], h5 = h[1*3+1], h7 = h[2*3+0], h8 = h[2*3+1];
	const double focalNumerator =	+ (h2*(h2 + h4) - h1*h5 + pow(h5,2))*(pow(h2,2) - h2*h4 + h5*(h1 + h5))*pow(h7,2) - (pow(h1,2)
					+ pow(h4,2))*(h1*h2 + h4*h5)*h7*h8 + (pow(h1,2) + pow(h4,2))*(pow(h1,2) - pow(h2,2) + pow(h4,2) - pow(h5,2))*pow(h8,2);
	const double focalDenominator = (pow(h2,2) + pow(h5,2))*pow(h7,4) - (h1*h2 + h4*h5)*pow(h7,3)*h8 - (pow(h2,2) + pow(h5,2))*pow(h7,2)*pow(h8,2) + (pow(h1,2) + pow(h4,2))*pow(h8,4);
	const double focal = sqrt(ABS(focalNumerator / focalDenominator))/2;

	std::cout << "Focal = " << focal << std::endl;

	K = QVMatrix::identity(3);
	K(2,2) = 1/focal;
	}

//
// This function decomposes the matrix for an homography, obtaining matrix R and vector T from the decomposition:
//	H = K*M3x3
// where
//	M3x3 = [ R1 | R2 | -R^t*C ]
// obtaining the essencial matrix M of size 3x4:
//		[	|	 ]
//	M4x4 =	[  Rot  | -R^t*C ]
//		[ _____	| ______ ]
//		[ 0 0 0	|    1   ]	
//
void GetExtrinsicCameraMatrixFromHomography(const QVMatrix &K, const QVMatrix &H, QVMatrix &M4x4)
	{
	// 1. Eliminate K matrix from H
	QVMatrix M3x3 = pseudoInverse(K)*H;

	// 2. Tricky: homogenize M3x3 such as the lenght of rotation vectors is the unit.
	//	2.2 Divide M3x3 by the mean of the square values for first two elements
	//	of the diagonal of matrix M3x3^t * M3x3
	M3x3 = M3x3 / (0.5 * M3x3.getCol(0).norm2() + 0.5 * M3x3.getCol(1).norm2());

	// 3. Compose matrix R and -R*C vector in final matrix M
	M4x4 = QVMatrix(4,4);

	M4x4.setCol(0, M3x3.getCol(0));
	M4x4.setCol(1, M3x3.getCol(1));
	M4x4.setCol(2, M3x3.getCol(0) ^ M3x3.getCol(1));
	M4x4.setCol(3, M3x3.getCol(2));
	M4x4(3,3) = 1;
	}
